/***********************************************************************************************************************
 *
 * blueBill Mobile - open source birdwatching
 * ==========================================
 *
 * Copyright (C) 2009, 2010 by Tidalwave s.a.s. (http://www.tidalwave.it)
 * http://bluebill.tidalwave.it/mobile/
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: EntityWithProperties.java,v 69134dbcc5fb 2010/07/11 16:36:12 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.semantic;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import it.tidalwave.util.As;
import it.tidalwave.util.DefaultIdentifiable;
import it.tidalwave.util.Id;
import it.tidalwave.util.Key;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.TypeSafeHashMap;
import it.tidalwave.util.TypeSafeMap;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 *
 **********************************************************************************************************************/
@Immutable
public class EntityWithProperties<E extends EntityWithProperties> extends AsSupport implements As, TypeSafeMap
  {    
    private final TypeSafeMap map;

    @Nonnull
    private final Id id;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public EntityWithProperties (final @Nonnull Id id)
      {
        this(id, Collections.<Key<?>, Object>emptyMap());
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public EntityWithProperties (final Object ... capabilities)
      {
        super(capabilities);
        this.map = new TypeSafeHashMap(Collections.<Key<?>, Object>emptyMap());
        id = null;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected EntityWithProperties (final @Nonnull Id id, final @Nonnull Map<Key<?>, Object> map)
      {
        super(new DefaultIdentifiable(id));
        this.id = id;
        this.map = new TypeSafeHashMap(map);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Id getId()
      {
        return id;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> T get (final @Nonnull Key<T> key)
      throws NotFoundException
      {
        return map.get(key);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean containsKey (final @Nonnull Key<?> item)
      {
        return map.containsKey(item);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Set<Key<?>> getKeys()
      {
        return map.getKeys();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnegative
    public int getSize()
      {
        return map.getSize();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Iterator<Object> iterator()
      {
        return map.iterator();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> E with (final @Nonnull Key<T> key, final @Nonnull T value)
      {
        final Map<Key<?>, Object> asMap = map.asMap();
        final Object oldValue = asMap.get(key);

        if (oldValue != null)
          {
            final Collection<T> list = (oldValue instanceof Collection<?>) ? (Collection<T>)oldValue
                                                                           : new ArrayList<T>();
            if (list != oldValue)
              {
                list.add((T)oldValue);
              }
            
            list.add(value);
            asMap.put(key, list);
          }
        else
          {
            asMap.put(key, value);
          }

        final E clone = clone(asMap);
        clone.capabilities.addAll(capabilities);
        
        return clone;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected E clone (final @Nonnull Map<Key<?>, Object> asMap)
      {
        return (E)new EntityWithProperties(id, asMap);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Map<Key<?>, Object> asMap()
      {
        throw new UnsupportedOperationException("Not supported yet.");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public String toString()
      {
        return String.format("%s@%x[%s - %s]", getClass().getSimpleName(), System.identityHashCode(this), id, map);
      }
  }
